/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.layers.samediff.testlayers;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import java.util.*;
import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.NoArgsConstructor;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.samediff.SDLayerParams;
import org.deeplearning4j.nn.conf.layers.samediff.SameDiffLayer;
import org.deeplearning4j.nn.conf.layers.samediff.SameDiffLayerUtils;
import org.deeplearning4j.nn.params.DefaultParamInitializer;
import org.deeplearning4j.nn.weights.WeightInitUtil;
import org.nd4j.autodiff.samediff.SDVariable;
import org.nd4j.autodiff.samediff.SameDiff;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.ndarray.INDArray;

@Data
@EqualsAndHashCode(
    callSuper = true,
    exclude = {"paramShapes"})
@NoArgsConstructor()
@JsonIgnoreProperties("paramShapes")
@SuperBuilder
public class SameDiffDense extends SameDiffLayer {

  private static final List<String> W_KEYS =
      Collections.singletonList(DefaultParamInitializer.WEIGHT_KEY);
  private static final List<String> B_KEYS =
      Collections.singletonList(DefaultParamInitializer.BIAS_KEY);
  private static final List<String> PARAM_KEYS =
      Arrays.asList(DefaultParamInitializer.WEIGHT_KEY, DefaultParamInitializer.BIAS_KEY);

  private final Map<String, long[]> paramShapes = new HashMap<>();

  private long nIn;
  private long nOut;
  private Activation activation;

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    return null;
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (override) {
      this.nIn = ((InputType.InputTypeFeedForward) inputType).getSize();
    }
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    return null;
  }

  @Override
  public void defineParameters(SDLayerParams params) {
    params.clear();
    params.addWeightParam(DefaultParamInitializer.WEIGHT_KEY, nIn, nOut);
    params.addBiasParam(DefaultParamInitializer.BIAS_KEY, 1, nOut);
  }

  @Override
  public void initializeParameters(Map<String, INDArray> params) {
    for (Map.Entry<String, INDArray> e : params.entrySet()) {
      if (paramWeightInit != null && paramWeightInit.containsKey(e.getKey())) {
        paramWeightInit.get(e.getKey()).init(nIn, nOut, e.getValue().shape(), 'c', e.getValue());
      } else {
        if (DefaultParamInitializer.BIAS_KEY.equals(e.getKey())) {
          e.getValue().assign(0.0);
        } else {
          // Normally use 'c' order, but use 'f' for direct comparison to DL4J
          // DenseLayerConfiguration
          WeightInitUtil.initWeights(
              nIn, nOut, new long[] {nIn, nOut}, weightInit, null, 'f', e.getValue());
        }
      }
    }
  }

  @Override
  public SDVariable defineLayer(
      SameDiff sd, SDVariable layerInput, Map<String, SDVariable> paramTable, SDVariable mask) {
    SDVariable weights = paramTable.get(DefaultParamInitializer.WEIGHT_KEY);
    SDVariable bias = paramTable.get(DefaultParamInitializer.BIAS_KEY);

    SDVariable mmul = sd.mmul("mmul", layerInput, weights);
    SDVariable z = mmul.add("z", bias);
    return activation.asSameDiff("out", sd, z);
  }

  @Override
  public void applyGlobalConfigToLayer(
      NeuralNetConfiguration.NeuralNetConfigurationBuilder globalConfig) {
    NeuralNetConfiguration clone = globalConfig.clone().build();
    if (activation == null) {
      activation = SameDiffLayerUtils.fromIActivation(clone.getActivation());
    }
  }

  public char paramReshapeOrder(String param) {
    // To match DL4J for easy comparison
    return 'f';
  }
}
