/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import net.brutex.ai.dnn.api.LayerType;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.activations.IActivation;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.learning.config.IUpdater;


@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(buildMethodName = "initBuild", builderMethodName = "innerBuilder")
public class ActivationLayer extends NoParamLayer {


  public static ActivationLayerBuilder<?, ?> builder(Activation activation) {
    return innerBuilder().activation(activation);
  }
  public static ActivationLayerBuilder<?, ?> builder(IActivation activation) {
    return innerBuilder().activation(activation);
  }


  public static ActivationLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }
  @Override
  public ActivationLayer clone() {
    ActivationLayer clone = (ActivationLayer) super.clone();
    return clone;
  }

  @Override
  public IUpdater getIUpdater() {
    return null;
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    this.setNetConfiguration(conf);
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);

    org.deeplearning4j.nn.layers.ActivationLayer ret =
        new org.deeplearning4j.nn.layers.ActivationLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input type: null for layer name \"" + getName() + "\"");
    }
    return inputType;
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    // No input preprocessor required for any input
    return null;
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    val actElementsPerEx = inputType.arrayElementsPerExample();

    return new LayerMemoryReport.Builder(name, ActivationLayer.class, inputType, inputType)
        .standardMemory(0, 0) // No params
        // During inference: modify input activation in-place
        // During  backprop: dup the input for later re-use
        .workingMemory(0, 0, 0, actElementsPerEx)
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching
        .build();
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    // No op
  }

  public static abstract class ActivationLayerBuilder<
          C extends ActivationLayer, B extends ActivationLayerBuilder<C, B>>
      extends NoParamLayer.NoParamLayerBuilder<C, B> {
    public C build() {
      C l = this.initBuild();
      l.initializeConstraints();
      return l;
    }
  }
}
