/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.api.ITraininableLayerConfiguration;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.GradientNormalization;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.Updater;
import org.deeplearning4j.nn.conf.distribution.Distribution;
import org.deeplearning4j.nn.weights.IWeightInit;
import org.deeplearning4j.nn.weights.WeightInit;
import org.deeplearning4j.nn.weights.WeightInitDistribution;
import org.deeplearning4j.util.NetworkUtils;
import org.jetbrains.annotations.NotNull;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.regularization.L1Regularization;
import org.nd4j.linalg.learning.regularization.L2Regularization;
import org.nd4j.linalg.learning.regularization.Regularization;
import org.nd4j.linalg.learning.regularization.WeightDecay;

/** A neural network layer. */
@EqualsAndHashCode(callSuper = true)
@NoArgsConstructor
@SuperBuilder
public abstract class BaseLayerConfiguration extends LayerConfiguration
    implements ITraininableLayerConfiguration, Serializable, Cloneable {


  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
   * regularization, etc). These constraints are applied at each iteration, after the parameters
   * have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all bias parameters of all layers
   */
  //@lombok.Builder.Default @Getter protected final List<LayerConstraint> biasConstraints = new ArrayList<>();
  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
   * regularization, etc). These constraints are applied at each iteration, after the parameters
   * have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all weight parameters of all layers
   */
  @lombok.Builder.Default @Getter
  protected final List<LayerConstraint> constrainWeights = new ArrayList<>();
  /** Weight initialization scheme to use, for initial weight values */
  @Getter @Setter
  protected IWeightInit weightInit;
  /** Bias initialization value, for layers with biases. Defaults to 0 */
  @Getter @Setter @Builder.Default
  protected double biasInit = 0.0;
  /** Gain initialization value, for layers with ILayer Normalization. Defaults to 1 */
  @Getter @Setter @Builder.Default
  protected double gainInit = 0.0;
  /** Regularization for the parameters (excluding biases). */
  @Builder.Default @Getter @Setter protected List<Regularization> regularization = new ArrayList<>();
  /** Regularization for the bias parameters only */
  @Builder.Default @Getter @Setter
  protected List<Regularization> regularizationBias = new ArrayList<>();
  /**
   * Gradient updater. For example, {@link org.nd4j.linalg.learning.config.Adam} or {@link
   * org.nd4j.linalg.learning.config.Nesterovs}
   */
  @Getter @Setter
  protected IUpdater updater;
  /**
   * Gradient updater configuration, for the biases only. If not set, biases will use the updater as
   * set by {@link #setUpdater(IUpdater)}
   */
  @Getter @Setter
  protected IUpdater biasUpdater;
  /**
   * Gradient normalization strategy. Used to specify gradient renormalization, gradient clipping
   * etc. Defaults to {@link GradientNormalization#None}.
   *
   * @see GradientNormalization
   */
  @Builder.Default
  protected @Getter @Setter GradientNormalization gradientNormalization =
      GradientNormalization.None; // Clipping, rescale based on l2 norm, etc
  /**
   * Threshold for gradient normalization, only used for GradientNormalization.ClipL2PerLayer,
   * GradientNormalization.ClipL2PerParamType, and
   * GradientNormalization.ClipElementWiseAbsoluteValue<br>
   * Not used otherwise.<br>
   * L2 threshold for first two types of clipping, or absolute value threshold for last type of
   * clipping.
   */
  @Builder.Default @Getter @Setter
  protected double gradientNormalizationThreshold =
      1.0; // Threshold for l2 and element-wise gradient clipping
@Getter @Setter
  private DataType dataType;

  /**
   * Reset the learning related configs of the layer to default. When instantiated with a global
   * neural network configuration the parameters specified in the neural network configuration will
   * be used. For internal use with the transfer learning API. Users should not have to call this
   * method directly.
   */
  public void resetLayerDefaultConfig() {
    // clear the learning related params for all layers in the origConf and set to defaults
    this.setIUpdater(null);
    this.setWeightInit(null);
    this.setBiasInit(Double.NaN);
    this.setGainInit(Double.NaN);
    this.regularization = null;
    this.regularizationBias = null;
    this.setGradientNormalization(GradientNormalization.None);
    this.setGradientNormalizationThreshold(1.0);
    this.updater = null;
    this.biasUpdater = null;
  }

  @Override
  public BaseLayerConfiguration clone() {
    BaseLayerConfiguration clone = (BaseLayerConfiguration) super.clone();
    if (regularization != null) {
      // Regularization fields are _usually_ thread safe and immutable, but let's clone to be sure
      clone.regularization = new ArrayList<>(regularization.size());
      for (Regularization r : regularization) {
        clone.regularization.add(r.clone());
      }
    }
    if (regularizationBias != null) {
      clone.regularizationBias = new ArrayList<>(regularizationBias.size());
      for (Regularization r : regularizationBias) {
        clone.regularizationBias.add(r.clone());
      }
    }
    return clone;
  }

  /**
   * Get the updater for the given parameter. Typically the same updater will be used for all
   * parameters, but this is not necessarily the case
   *
   * @param paramName Parameter name
   * @return IUpdater for the parameter
   */
  @Override
  public IUpdater getUpdaterByParam(String paramName) {
    if (biasUpdater != null && initializer().isBiasParam(this, paramName)) {
      return biasUpdater;
    }
    return updater;
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    if (initializer().isWeightParam(this, paramName)) {
      return regularization;
    } else if (initializer().isBiasParam(this, paramName)) {
      return regularizationBias;
    }
    return null;
  }

  /**
   * Inherit setting from neural network for those settings, that are not already set or do have a
   * layer(type) specific default.
   *
   * @param conf the neural net configration to inherit parameters from
   */
  @Override
  public void runInheritance(@NotNull NeuralNetConfiguration conf) {
    super.runInheritance(conf);
    if (this.biasUpdater == null) this.biasUpdater = conf.getBiasUpdater();
    if (this.updater == null) this.updater = conf.getUpdater();
    if (this.regularizationBias == null) this.regularizationBias = conf.getRegularizationBias();
    if (this.regularization == null) this.regularization = conf.getRegularization();
    if (this.gradientNormalization == null)
      this.gradientNormalization = conf.getGradientNormalization();
    // if(this.weightInit == null) this.weightInit = conf.getWeightInit();

  }

  public static abstract class BaseLayerConfigurationBuilder<
          C extends BaseLayerConfiguration, B extends BaseLayerConfigurationBuilder<C, B>>
      extends LayerConfigurationBuilder<C, B> {

public B updater(Updater upd) {
  this.updater = upd.getIUpdaterWithDefaultConfig();
  return self();
}


    public B updater(IUpdater upd) {
      this.updater = upd;
      return self();
    }

    /**
     * Set weight initialization scheme to random sampling via the specified distribution.
     * Equivalent to: {@code .weightInit(new WeightInitDistribution(distribution))}
     *
     * @param dist Distribution to use for weight initialization
     */
    @Deprecated
    public B dist(Distribution dist) {
      this.weightInit = new WeightInitDistribution(dist);
      return self();
    }

    /**
     * Weight initialization scheme to use, for initial weight values
     *
     * @see WeightInit
     */
    public B weightInit(WeightInit weightInit) {
      if (weightInit == WeightInit.DISTRIBUTION) {
        throw new UnsupportedOperationException(
            "Not supported!, Use weightInit(Distribution distribution) instead!");
      }
      this.weightInit = weightInit.getWeightInitFunction();
      return self();
    }

    public B weightInit(IWeightInit weightInit) {
      if (weightInit.enumValue() == WeightInit.DISTRIBUTION) {
        throw new UnsupportedOperationException(
            "Not supported!, Use weightInit(Distribution distribution) instead!");
      }
      this.weightInit = weightInit.enumValue().getWeightInitFunction();
      return self();
    }

    public B weightInit(Distribution dist) {
      this.weightInit = new WeightInitDistribution(dist);
      return self();
    }

    /**
     * L1 regularization coefficient (weights only). Use {@link #l1Bias(double)} to configure the l1
     * regularization coefficient for the bias.
     */
    public B l1(double l1) {
      // Check if existing L1 exists; if so, replace it
      NetworkUtils.removeInstances(regularization$value, L1Regularization.class);
      if (l1 > 0.0) {
        this.regularization$value.add(new L1Regularization(l1));
      }
      regularization$set = true;
      return self();
    }

    /**
     * L2 regularization coefficient (weights only). Use {@link #l2Bias(double)} to configure the l2
     * regularization coefficient for the bias.<br>
     * <b>Note</b>: Generally, {@link org.nd4j.linalg.learning.regularization.WeightDecay} (set via
     * {@link #weightDecay(double,boolean)} should be preferred to L2 regularization. See {@link
     * org.nd4j.linalg.learning.regularization.WeightDecay} javadoc for further details.<br>
     */
    public B l2(double l2) {
      // Check if existing L2 exists; if so, replace it. Also remove weight decay - it doesn't make
      // sense to use both
      NetworkUtils.removeInstances(this.regularization$value, L2Regularization.class);
      if (l2 > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularization$value,
            WeightDecay.class,
            "WeightDecay regularization removed: incompatible with added L2 regularization");
        this.regularization$value.add(new L2Regularization(l2));
      }
      regularization$set = true;
      return self();
    }

    /** L1 regularization coefficient for the bias. Default: 0. See also {@link #l1(double)} */
    public B l1Bias(double l1Bias) {
      NetworkUtils.removeInstances(this.regularizationBias$value, L1Regularization.class);
      if (l1Bias > 0.0) {
        this.regularizationBias$value.add(new L1Regularization(l1Bias));
      }
      regularizationBias$set = true;
      return self();
    }

    /**
     * L2 regularization coefficient for the bias. Default: 0. See also {@link #l2(double)}<br>
     * <b>Note</b>: Generally, {@link WeightDecay} (set via {@link #weightDecayBias(double,boolean)}
     * should be preferred to L2 regularization. See {@link WeightDecay} javadoc for further
     * details.<br>
     */
    public B l2Bias(double l2Bias) {
      NetworkUtils.removeInstances(this.regularizationBias$value, L2Regularization.class);
      if (l2Bias > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularizationBias$value,
            WeightDecay.class,
            "WeightDecay regularization removed: incompatible with added L2 regularization");
        this.regularizationBias$value.add(new L2Regularization(l2Bias));
      }
      regularizationBias$set = true;
      return self();
    }

    /**
     * Add weight decay regularization for the network parameters (excluding biases).<br>
     * This applies weight decay <i>with</i> multiplying the learning rate - see {@link WeightDecay}
     * for more details.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @see #weightDecay(double, boolean)
     */
    public B weightDecay(double coefficient) {
      return weightDecay(coefficient, true);
    }

    /**
     * Add weight decay regularization for the network parameters (excluding biases). See {@link
     * WeightDecay} for more details.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @param applyLR Whether the learning rate should be multiplied in when performing weight decay
     *     updates. See {@link WeightDecay} for more details.
     * @see #weightDecay(double, boolean)
     */
    public B weightDecay(double coefficient, boolean applyLR) {
      // Check if existing weight decay if it exists; if so, replace it. Also remove L2 - it doesn't
      // make sense to use both
      NetworkUtils.removeInstances(this.regularization$value, WeightDecay.class);
      if (coefficient > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularization$value,
            L2Regularization.class,
            "L2 regularization removed: incompatible with added WeightDecay regularization");
        this.regularization$value.add(new WeightDecay(coefficient, applyLR));
      }
      regularization$set = true;
      return self();
    }

    /**
     * Set constraints to be applied to all layers. Default: no constraints.<br>
     * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
     * regularization, etc). These constraints are applied at each iteration, after the parameters
     * have been updated.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     *
     * @param constraints Constraints to apply to all weight parameters of all layers
     */
    public B constrainWeights(LayerConstraint... constraints) {
      constrainWeights$value = Arrays.asList(constraints);
      constrainWeights$set = true;
      return self();
    }

    /**
     * Weight decay for the biases only - see {@link #weightDecay(double)} for more details. This
     * applies weight decay <i>with</i> multiplying the learning rate.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @see #weightDecayBias(double, boolean)
     */
    public B weightDecayBias(double coefficient) {
      return weightDecayBias(coefficient, true);
    }



    /**
     * Weight decay for the biases only - see {@link #weightDecay(double)} for more details<br>
     *
     * @param coefficient Weight decay regularization coefficient
     */
    public B weightDecayBias(double coefficient, boolean applyLR) {
      // Check if existing weight decay if it exists; if so, replace it. Also remove L2 - it doesn't
      // make sense to use both
      NetworkUtils.removeInstances(this.regularizationBias$value, WeightDecay.class);
      if (coefficient > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            this.regularizationBias$value,
            L2Regularization.class,
            "L2 regularization removed: incompatible with added WeightDecay regularization");
        this.regularizationBias$value.add(new WeightDecay(coefficient, applyLR));
      }
      regularizationBias$set = true;
      return self();
    }
  }
}
