/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Arrays;
import java.util.List;
import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.weights.IWeightInit;


@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@SuperBuilder
public abstract class BaseRecurrentLayer extends FeedForwardLayer {

    /**
     * Set the weight initialization for the recurrent weights. Not that if this is not set explicitly, the same
     * weight initialization as the layer input weights is also used for the recurrent weights.
     *
     */
    @Getter
    protected IWeightInit weightInitRecurrent;
    /**
     * Set the format of data expected by the RNN. NCW = [miniBatchSize, size, timeSeriesLength],
     * NWC = [miniBatchSize, timeSeriesLength, size]. Defaults to NCW.
     */
    @Builder.Default @Getter @Setter
    protected RNNFormat dataFormat = RNNFormat.NCW;
    /**
     * Set constraints to be applied to the RNN recurrent weight parameters of this layer. Default: no
     * constraints.<br> Constraints can be used to enforce certain conditions (non-negativity of parameters,
     * max-norm regularization, etc). These constraints are applied at each iteration, after the parameters have
     * been updated.
     */
    @Getter
    protected List<LayerConstraint> recurrentConstraints;
    /**
     * Set constraints to be applied to the RNN input weight parameters of this layer. Default: no constraints.<br>
     * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm regularization,
     * etc). These constraints are applied at each iteration, after the parameters have been updated.
     *
     */
    @Getter
    protected List<LayerConstraint> inputWeightConstraints;

    @Override
    public InputType getOutputType(int layerIndex, InputType inputType) {
        if (inputType == null || inputType.getType() != InputType.Type.RNN) {
            throw new IllegalStateException("Invalid input for RNN layer (layer index = " + layerIndex
                    + ", layer name = \"" + getName() + "\"): expect RNN input type with size > 0. Got: "
                    + inputType);
        }

        InputType.InputTypeRecurrent itr = (InputType.InputTypeRecurrent) inputType;

        return InputType.recurrent(nOut, itr.getTimeSeriesLength(), itr.getFormat());
    }

    @Override
    public void setNIn(InputType inputType, boolean override) {
        if (inputType == null || inputType.getType() != InputType.Type.RNN) {
            throw new IllegalStateException("Invalid input for RNN layer (layer name = \"" + getName()
                    + "\"): expect RNN input type with size > 0. Got: " + inputType);
        }

        InputType.InputTypeRecurrent r = (InputType.InputTypeRecurrent) inputType;
        if (nIn <= 0 || override) {
            this.nIn = r.getSize();
        }

        if(dataFormat == null || override)
            this.dataFormat = r.getFormat();
    }

    @Override
    public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
        return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, dataFormat, getName());
    }

    public static abstract class BaseRecurrentLayerBuilder<C extends BaseRecurrentLayer, B extends BaseRecurrentLayerBuilder<C, B>>
    extends FeedForwardLayerBuilder<C, B> {
        /**
         * Set constraints to be applied to the RNN recurrent weight parameters of this layer. Default: no
         * constraints.<br> Constraints can be used to enforce certain conditions (non-negativity of parameters,
         * max-norm regularization, etc). These constraints are applied at each iteration, after the parameters have
         * been updated.
         *
         * @param constraints Constraints to apply to the recurrent weight parameters of this layer
         */
        public B constrainRecurrent(LayerConstraint... constraints) {
            this.recurrentConstraints(Arrays.asList(constraints));
            return self();
        }

        /**
         * Set constraints to be applied to the RNN input weight parameters of this layer. Default: no constraints.<br>
         * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm regularization,
         * etc). These constraints are applied at each iteration, after the parameters have been updated.
         *
         * @param constraints Constraints to apply to the input weight parameters of this layer
         */
        public B constrainInputWeights(LayerConstraint... constraints) {
            this.inputWeightConstraints(Arrays.asList(constraints));
            return self();
        }
    }
}
