/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import com.fasterxml.jackson.annotation.JsonIgnore;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import net.brutex.ai.dnn.api.LayerType;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.*;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.ConvolutionParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ConvolutionUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

/**
 * ConvolutionLayer nIn in the input layer is the number of channels nOut is the number of filters
 * to be used in the net or in other words the channels The builder specifies the filter/kernel
 * size, the stride and padding The pooling layer takes the kernel size
 */
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(builderMethodName = "innerBuilder", buildMethodName = "initBuild")
public class ConvolutionLayer extends FeedForwardLayer {
  /**
   * Size of the convolution rows/columns
   *
   * @param kernelSize the height and width of the kernel
   */
  private @Builder.Default @Getter @Setter int[] kernelSize = new int[] {5, 5}; // Square filter
  /** If true (default): include bias parameters in the model. False: no bias. */
  @Builder.Default @Getter @Setter private boolean hasBias = true;
  /**
   * Set the convolution mode for the Convolution layer. See {@link ConvolutionMode} for more
   * details Default is {@link ConvolutionMode}.Truncate.
   */
  @Builder.Default @Getter @Setter
  private ConvolutionMode convolutionMode = ConvolutionMode.Truncate;

  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default @Getter @Setter
  private CNN2DFormat convFormat =
      CNN2DFormat.NCHW; // default value for legacy serialization reasons

  /**
   * Kernel dilation. Default: {1, 1}, which is standard convolutions. Used for implementing dilated
   * convolutions, which are also known as atrous convolutions.
   *
   * <p>For more details, see: <a href="https://arxiv.org/abs/1511.07122">Yu and Koltun (2014)</a>
   * and <a href="https://arxiv.org/abs/1412.7062">Chen et al. (2014)</a>, as well as <a
   * href="http://deeplearning.net/software/theano/tutorial/conv_arithmetic.html#dilated-convolutions">
   * http://deeplearning.net/software/theano/tutorial/conv_arithmetic.html#dilated-convolutions</a>
   * <br>
   */
  @Getter @Setter private @Builder.Default int[] dilation = new int[] {1, 1};
  /** Default is 2. Down-sample by a factor of 2 */
  @Getter @Setter private @Builder.Default int[] stride = new int[] {1, 1};

  @Getter @Setter private @Builder.Default int[] padding = new int[] {0, 0};
  /**
   * When using CuDNN and an error is encountered, should fallback to the non-CuDNN implementatation
   * be allowed? If set to false, an exception in CuDNN will be propagated back to the user. If
   * false, the built-in (non-CuDNN) implementation for ConvolutionLayer will be used
   */
  @Getter @Setter @Builder.Default private boolean cudnnAllowFallback = true;

  /** Defaults to "PREFER_FASTEST", but "NO_WORKSPACE" uses less memory. */
  @Getter @Setter @Builder.Default private AlgoMode cudnnAlgoMode = AlgoMode.PREFER_FASTEST;

  @Getter @Setter private FwdAlgo cudnnFwdAlgo;
  @Getter @Setter private BwdFilterAlgo cudnnBwdFilterAlgo;
  @Getter @Setter private BwdDataAlgo cudnnBwdDataAlgo;
  @Getter @Setter @Builder.Default private int convolutionDim = 2; // 2D convolution by default
  /** Causal convolution - allowed for 1D only */
  @Builder.Default private boolean allowCausal = false;

  @Builder.Default @JsonIgnore @EqualsAndHashCode.Exclude @Getter @Setter
  private boolean defaultValueOverriden = false;

  public static ConvolutionLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static ConvolutionLayerBuilder<?, ?> builder(int... kernelSize) {
    return innerBuilder().kernelSize(kernelSize);
  }

  public static ConvolutionLayerBuilder<?, ?> builder(int[] kernelSize, int[] stride) {
    return innerBuilder().kernelSize(kernelSize).stride(stride);
  }

  public static ConvolutionLayerBuilder<?, ?> builder(
      int[] kernelSize, int[] stride, int[] padding) {
    return innerBuilder().kernelSize(kernelSize).stride(stride).padding(padding);
  }

  public boolean hasBias() {
    return hasBias;
  }

  @Override
  public ConvolutionLayer clone() {
    ConvolutionLayer clone = (ConvolutionLayer) super.clone();
    if (clone.kernelSize != null) {
      clone.kernelSize = clone.kernelSize.clone();
    }
    if (clone.stride != null) {
      clone.stride = clone.stride.clone();
    }
    if (clone.padding != null) {
      clone.padding = clone.padding.clone();
    }
    return clone;
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    setNetConfiguration(conf);
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();

    LayerValidation.assertNInNOutSet(
        "ConvolutionLayer", getName(), layerIndex, getNIn(), getNOut());

    org.deeplearning4j.nn.layers.convolution.ConvolutionLayer ret =
        new org.deeplearning4j.nn.layers.convolution.ConvolutionLayer(lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return ConvolutionParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for Convolution layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }

    return InputTypeUtil.getOutputTypeCnnLayers(
        inputType,
        kernelSize,
        stride,
        padding,
        dilation,
        convolutionMode,
        nOut,
        layerIndex,
        getName(),
        convFormat,
        ConvolutionLayer.class);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for Convolution layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }

    if (!defaultValueOverriden || nIn <= 0 || override) {
      InputType.InputTypeConvolutional c = (InputType.InputTypeConvolutional) inputType;
      this.nIn = c.getChannels();
      this.convFormat = ((InputType.InputTypeConvolutional) inputType).getFormat();
    }

    if (convFormat == null || override)
      this.convFormat = ((InputType.InputTypeConvolutional) inputType).getFormat();
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Convolution layer (layer name=\"" + getName() + "\"): input is null");
    }

    return InputTypeUtil.getPreProcessorForInputTypeCnnLayers(inputType, getName());
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    val paramSize = initializer().numParams(this);
    val updaterStateSize = (int) getIUpdater().stateSize(paramSize);

    InputType.InputTypeConvolutional c = (InputType.InputTypeConvolutional) inputType;
    InputType.InputTypeConvolutional outputType =
        (InputType.InputTypeConvolutional) getOutputType(-1, inputType);

    // TODO convolution helper memory use... (CuDNN etc)

    // During forward pass: im2col array, mmul (result activations), in-place broadcast add
    val im2colSizePerEx =
        c.getChannels()
            * outputType.getHeight()
            * outputType.getWidth()
            * kernelSize[0]
            * kernelSize[1];

    // During training: have im2col array, in-place gradient calculation, then epsilons...
    // But: im2col array may be cached...
    Map<CacheMode, Long> trainWorkingMemoryPerEx = new HashMap<>();
    Map<CacheMode, Long> cachedPerEx = new HashMap<>();

    // During backprop: im2col array for forward pass (possibly cached) + the epsilon6d array
    // required to calculate
    // the 4d epsilons (equal size to input)
    // Note that the eps6d array is same size as im2col
    for (CacheMode cm : CacheMode.values()) {
      long trainWorkingSizePerEx;
      long cacheMemSizePerEx = 0;
      if (cm == CacheMode.NONE) {
        trainWorkingSizePerEx = 2 * im2colSizePerEx;
      } else {
        // im2col is cached, but epsNext2d/eps6d is not
        cacheMemSizePerEx = im2colSizePerEx;
        trainWorkingSizePerEx = im2colSizePerEx;
      }

      if (getDropOut() != null) {
        // Dup on the input before dropout, but only for training
        trainWorkingSizePerEx += inputType.arrayElementsPerExample();
      }

      trainWorkingMemoryPerEx.put(cm, trainWorkingSizePerEx);
      cachedPerEx.put(cm, cacheMemSizePerEx);
    }

    return new LayerMemoryReport.Builder(name, ConvolutionLayer.class, inputType, outputType)
        .standardMemory(paramSize, updaterStateSize)
        // im2col caching -> only variable size caching
        .workingMemory(
            0, im2colSizePerEx, MemoryReport.CACHE_MODE_ALL_ZEROS, trainWorkingMemoryPerEx)
        .cacheMemory(MemoryReport.CACHE_MODE_ALL_ZEROS, cachedPerEx)
        .build();
  }
  /**
   * The "PREFER_FASTEST" mode will pick the fastest algorithm for the specified parameters from the
   * {@link FwdAlgo}, {@link BwdFilterAlgo}, and {@link BwdDataAlgo} lists, but they may be very
   * memory intensive, so if weird errors occur when using cuDNN, please try the "NO_WORKSPACE"
   * mode. Alternatively, it is possible to specify the algorithm manually by setting the
   * "USER_SPECIFIED" mode, but this is not recommended.
   *
   * <p>Note: Currently only supported with cuDNN.
   */
  public enum AlgoMode {
    NO_WORKSPACE,
    PREFER_FASTEST,
    USER_SPECIFIED
  }

  /**
   * The forward algorithm to use when {@link AlgoMode} is set to "USER_SPECIFIED".
   *
   * <p>Note: Currently only supported with cuDNN.
   */
  public enum FwdAlgo {
    IMPLICIT_GEMM,
    IMPLICIT_PRECOMP_GEMM,
    GEMM,
    DIRECT,
    FFT,
    FFT_TILING,
    WINOGRAD,
    WINOGRAD_NONFUSED,
    COUNT
  }

  /**
   * The backward filter algorithm to use when {@link AlgoMode} is set to "USER_SPECIFIED".
   *
   * <p>Note: Currently only supported with cuDNN.
   */
  public enum BwdFilterAlgo {
    ALGO_0,
    ALGO_1,
    FFT,
    ALGO_3,
    WINOGRAD,
    WINOGRAD_NONFUSED,
    FFT_TILING,
    COUNT
  }

  /**
   * The backward data algorithm to use when {@link AlgoMode} is set to "USER_SPECIFIED".
   *
   * <p>Note: Currently only supported with cuDNN.
   */
  public enum BwdDataAlgo {
    ALGO_0,
    ALGO_1,
    FFT,
    FFT_TILING,
    WINOGRAD,
    WINOGRAD_NONFUSED,
    COUNT
  }

  public abstract static class ConvolutionLayerBuilder<
          C extends ConvolutionLayer, B extends ConvolutionLayerBuilder<C, B>>
      extends FeedForwardLayerBuilder<C, B> {

    /**
     * Set kernel size for 3D convolutions in (depth, height, width) order
     *
     * @param kernelSize kernel size
     */
    public B kernelSize(int... kernelSize) {
      this.kernelSize$value = ValidationUtils.validate2NonNegative(kernelSize, false, "kernelSize");
      this.kernelSize$set = true;
      return self();
    }
    /**
     * Set stride size for 3D convolutions in (depth, height, width) order
     *
     * @param stride kernel size
     */
    public B stride(int... stride) {
      this.stride$value = ValidationUtils.validate2NonNegative(stride, false, "stride");
      this.stride$set = true;
      return self();
    }

    /**
     * Set padding size for 3D convolutions in (depth, height, width) order
     *
     * @param padding kernel size
     */
    public B padding(int... padding) {
      this.padding$value = ValidationUtils.validate2NonNegative(padding, false, "padding");
      this.padding$set = true;
      return self();
    }
    /**
     * Set dilation size for 3D convolutions in (depth, height, width) order
     *
     * @param dilation kernel size
     */
    public B dilation(int... dilation) {
      this.dilation$value = ValidationUtils.validate2NonNegative(dilation, false, "dilation");
      this.dilation$set = true;
      return self();
    }

    public C build() {
      ConvolutionUtils.validateConvolutionModePadding(convolutionMode$value, padding$value);
      ConvolutionUtils.validateCnnKernelStridePadding(
          kernelSize$value, stride$value, padding$value);

      if (kernelSize$value.length != convolutionDim$value) {
        throw new IllegalArgumentException(
            "Kernel argument should be a "
                + convolutionDim$value
                + "d array, got "
                + Arrays.toString(kernelSize$value));
      }

      if (stride$value.length != convolutionDim$value) {
        throw new IllegalArgumentException(
            "Strides argument should be a "
                + convolutionDim$value
                + "d array, got "
                + Arrays.toString(stride$value));
      }

      if (padding$value.length != convolutionDim$value) {
        throw new IllegalArgumentException(
            "Padding argument should be a "
                + convolutionDim$value
                + "d array, got "
                + Arrays.toString(padding$value));
      }

      if (dilation$value.length != convolutionDim$value) {
        throw new IllegalArgumentException(
            "Dilation argument should be a "
                + convolutionDim$value
                + "d array, got "
                + Arrays.toString(dilation$value));
      }

      C l = initBuild();
      l.setType(LayerType.CONV);
      l.initializeConstraints();
      return l;
    }

    /**
     * When using CuDNN or MKLDNN and an error is encountered, should fallback to the non-helper
     * implementation be allowed? If set to false, an exception in the helper will be propagated
     * back to the user. If true, the built-in (non-MKL/CuDNN) implementation for ConvolutionLayer
     * will be used
     *
     * @param allowFallback Whether fallback to non-CuDNN implementation should be used
     */
    public B helperAllowFallback(boolean allowFallback) {
      this.cudnnAllowFallback$value = allowFallback;
      this.cudnnAllowFallback$set = true;
      return self();
    }
  }
}
