/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import net.brutex.ai.dnn.api.LayerType;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.dropout.Dropout;
import org.deeplearning4j.nn.conf.dropout.IDropout;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.learning.regularization.Regularization;

/**
 * Create a dropout layer with standard {@link Dropout}, with the specified probability of retaining
 * the input activation. See {@link Dropout} for the full details
 */
@Data

@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(builderMethodName = "innerBuilder")
public class DropoutLayer extends FeedForwardLayer {


  public static DropoutLayerBuilder<?,?> builder() {
    return innerBuilder();
  }

  public static DropoutLayerBuilder<?,?> builder(double dropout ) {
    return innerBuilder()
            .dropOut(dropout);
  }
  public static DropoutLayerBuilder<?,?> builder(IDropout dropout ) {
    return innerBuilder()
            .dropOut(dropout);
  }
  @Override
  public DropoutLayer clone() {
    return (DropoutLayer) super.clone();
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    setNetConfiguration(conf);

    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.DropoutLayer ret =
        new org.deeplearning4j.nn.layers.DropoutLayer(lconf, networkDataType);

    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return EmptyParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input type: null for layer name \"" + getName() + "\"");
    }
    return inputType;
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    // No op: dropout layer doesn't have a fixed nIn value
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    // No input preprocessor required; dropout applies to any input type
    return null;
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    // Not applicable
    return null;
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    throw new UnsupportedOperationException("Dropout layer does not contain parameters");
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    val actElementsPerEx = inputType.arrayElementsPerExample();
    // During inference: not applied. During  backprop: dup the input, in case it's used elsewhere
    // But: this will be counted in the activations
    // (technically inference memory is over-estimated as a result)

    return new LayerMemoryReport.Builder(name, DropoutLayer.class, inputType, inputType)
        .standardMemory(0, 0) // No params
        .workingMemory(0, 0, 0, 0) // No working mem, other than activations etc
        .cacheMemory(
            MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) // No caching
        .build();
  }

  public static abstract class DropoutLayerBuilder<
          C extends DropoutLayer, B extends DropoutLayerBuilder<C, B>>
      extends FeedForwardLayerBuilder<C, B> {

    public B dropout(IDropout dropout) {
      super.dropOut(dropout);
      return self();
    }
  }
}
