/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import java.util.*;
import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.conf.ConvolutionMode;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.samediff.SDLayerParams;
import org.deeplearning4j.nn.conf.layers.samediff.SameDiffLayer;
import org.deeplearning4j.nn.conf.layers.samediff.SameDiffLayerUtils;
import org.deeplearning4j.nn.params.ConvolutionParamInitializer;
import org.deeplearning4j.nn.weights.WeightInitUtil;
import org.deeplearning4j.util.Convolution1DUtils;
import org.nd4j.autodiff.samediff.SDIndex;
import org.nd4j.autodiff.samediff.SDVariable;
import org.nd4j.autodiff.samediff.SameDiff;
import org.nd4j.common.base.Preconditions;
import org.nd4j.enums.PadMode;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.memory.MemoryWorkspace;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.factory.Nd4j;

@Data
@EqualsAndHashCode(callSuper = true)
@JsonIgnoreProperties({"paramShapes"})
@SuperBuilder(buildMethodName = "initBuild")
public class LocallyConnected1D extends SameDiffLayer {

  public static abstract class LocallyConnected1DBuilder<C extends LocallyConnected1D, B extends LocallyConnected1DBuilder<C, B>> extends
  SameDiffLayerBuilder<C, B> {
    public C build() {
      Convolution1DUtils.validateConvolutionModePadding(convolutionMode$value, padding$value);
      Convolution1DUtils.validateCnn1DKernelStridePadding(kernelSize$value, stride$value, padding$value);
      C l = initBuild();
    return l;
    }
  }

  private static final List<String> WEIGHT_KEYS =
      Collections.singletonList(ConvolutionParamInitializer.WEIGHT_KEY);
  private static final List<String> BIAS_KEYS =
      Collections.singletonList(ConvolutionParamInitializer.BIAS_KEY);
  private static final List<String> PARAM_KEYS =
      Arrays.asList(ConvolutionParamInitializer.BIAS_KEY, ConvolutionParamInitializer.WEIGHT_KEY);
  /**
   * @param nIn Number of inputs to the layer (input size)
   */
  private long nIn;
  /**
   * @param nOut Number of outputs (output size)
   */
  private long nOut;
  /**
   * @param activation Activation function for the layer
   */
  @Builder.Default private Activation activation = Activation.TANH;
  /** Kernel size for the layer */
  @Builder.Default private int kernelSize = 2;
  /** Stride for the layer */
  @Builder.Default private int stride = 1;
  /** Padding for the layer. Not used if {@link ConvolutionMode#Same} is set */
  @Builder.Default private int padding = 0;

  private int paddingR; // Right/bottom padding
  /** Convolution mode for the layer. See {@link ConvolutionMode} for details */
  @Builder.Default private ConvolutionMode convolutionMode = ConvolutionMode.Same;

  /** Dilation for the layer */
  @Builder.Default private int dilation = 1;

  /** If true (default is false) the layer will have a bias */
  @Builder.Default private boolean hasBias = true;

  private int inputSize;
  private int outputSize;
  private int featureDim;

  public void computeOutputSize() {
    int nIn = (int) getNIn();
    if (inputSize == 0) {
      throw new IllegalArgumentException("Input size has to be set for Locally connected layers");
    }
    int[] inputShape = new int[] {1, nIn, inputSize};
    INDArray dummyInputForShapeInference = Nd4j.ones(inputShape);

    if (convolutionMode == ConvolutionMode.Same) {
      this.outputSize =
          Convolution1DUtils.getOutputSize(
              dummyInputForShapeInference, kernelSize, stride, 0, convolutionMode, dilation);
      this.padding =
          Convolution1DUtils.getSameModeTopLeftPadding(
              outputSize, inputSize, kernelSize, stride, dilation);
      this.paddingR =
          Convolution1DUtils.getSameModeBottomRightPadding(
              outputSize, inputSize, kernelSize, stride, dilation);
    } else {
      this.outputSize =
          Convolution1DUtils.getOutputSize(
              dummyInputForShapeInference, kernelSize, stride, padding, convolutionMode, dilation);
    }
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalArgumentException(
          "Provided input type for locally connected 1D layers has to be "
              + "of CNN1D/RNN type, got: "
              + inputType);
    }
    // dynamically compute input size from input type
    InputType.InputTypeRecurrent rnnType = (InputType.InputTypeRecurrent) inputType;
    this.inputSize = (int) rnnType.getTimeSeriesLength();
    computeOutputSize();

    return InputTypeUtil.getOutputTypeCnn1DLayers(
        inputType,
        kernelSize,
        stride,
        padding,
        1,
        convolutionMode,
        nOut,
        layerIndex,
        getName(),
        LocallyConnected1D.class);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (nIn <= 0 || override) {
      InputType.InputTypeRecurrent c = (InputType.InputTypeRecurrent) inputType;
      this.nIn = c.getSize();
    }
    if (featureDim <= 0 || override) {
      InputType.InputTypeRecurrent c = (InputType.InputTypeRecurrent) inputType;
      this.featureDim = kernelSize * (int) c.getSize();
    }
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, RNNFormat.NCW, getName());
  }

  @Override
  public void defineParameters(SDLayerParams params) {
    Preconditions.checkState(
        featureDim > 0, "Cannot initialize layer: Feature dimension is set to %s", featureDim);
    params.clear();
    val weightsShape = new long[] {outputSize, featureDim, nOut};
    params.addWeightParam(ConvolutionParamInitializer.WEIGHT_KEY, weightsShape);
    if (hasBias) {
      val biasShape = new long[] {nOut};
      params.addBiasParam(ConvolutionParamInitializer.BIAS_KEY, biasShape);
    }
  }

  @Override
  public void initializeParameters(Map<String, INDArray> params) {
    try (MemoryWorkspace ws = Nd4j.getWorkspaceManager().scopeOutOfWorkspaces()) {
      for (Map.Entry<String, INDArray> e : params.entrySet()) {
        if (ConvolutionParamInitializer.BIAS_KEY.equals(e.getKey())) {
          e.getValue().assign(0);
        } else {
          double fanIn = nIn * kernelSize;
          double fanOut = nOut * kernelSize / ((double) stride);
          WeightInitUtil.initWeights(
              fanIn, fanOut, e.getValue().shape(), weightInit, null, 'c', e.getValue());
        }
      }
    }
  }

  @Override
  public SDVariable defineLayer(
      SameDiff sameDiff,
      SDVariable layerInput,
      Map<String, SDVariable> paramTable,
      SDVariable mask) {
    SDVariable w =
        paramTable.get(ConvolutionParamInitializer.WEIGHT_KEY); // (outH, featureDim, nOut)

    int outH = outputSize;
    int sH = stride;
    int kH = kernelSize;

    if (padding > 0 || (convolutionMode == ConvolutionMode.Same && paddingR > 0)) {
      // Note: for same mode, bottom/right padding can be 1 more than top/left padding
      // NCW format.
      if (convolutionMode == ConvolutionMode.Same) {
        layerInput =
            sameDiff
                .nn()
                .pad(
                    layerInput,
                    sameDiff.constant(
                        Nd4j.createFromArray(new int[][] {{0, 0}, {0, 0}, {padding, paddingR}})),
                    PadMode.CONSTANT,
                    0);
      } else {
        layerInput =
            sameDiff
                .nn()
                .pad(
                    layerInput,
                    sameDiff.constant(
                        Nd4j.createFromArray(new int[][] {{0, 0}, {0, 0}, {padding, padding}})),
                    PadMode.CONSTANT,
                    0);
      }
    }

    SDVariable[] inputArray = new SDVariable[outH];
    for (int i = 0; i < outH; i++) {
      SDVariable slice =
          layerInput.get(
              SDIndex.all(), // miniBatch
              SDIndex.all(), // nIn
              SDIndex.interval(i * sH, i * sH + kH) // kernel
              );
      inputArray[i] = sameDiff.reshape(slice, 1, -1, featureDim);
    }
    SDVariable concatOutput = sameDiff.concat(0, inputArray); // (outH, miniBatch, featureDim)

    SDVariable mmulResult = sameDiff.mmul(concatOutput, w); // (outH, miniBatch, nOut)

    SDVariable result = sameDiff.permute(mmulResult, 1, 2, 0); // (miniBatch, nOut, outH)

    if (hasBias) {
      SDVariable b = paramTable.get(ConvolutionParamInitializer.BIAS_KEY);
      SDVariable biasAddedResult = sameDiff.nn().biasAdd(result, b, true);
      return activation.asSameDiff("out", sameDiff, biasAddedResult);
    } else {
      return activation.asSameDiff("out", sameDiff, result);
    }
  }

  @Override
  public void applyGlobalConfigToLayer(
      NeuralNetConfiguration.NeuralNetConfigurationBuilder globalConfig) {
    NeuralNetConfiguration global_conf = globalConfig.build();
    if (activation == null) {
      activation = SameDiffLayerUtils.fromIActivation(global_conf.getActivation());
    }
    if (convolutionMode == null) {
      convolutionMode = global_conf.getConvolutionMode();
    }
  }
}
