/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Map;
import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.conf.ConvolutionMode;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.inputs.InputType.InputTypeConvolutional;
import org.deeplearning4j.nn.conf.inputs.InputType.Type;
import org.deeplearning4j.nn.conf.layers.samediff.SDLayerParams;
import org.deeplearning4j.nn.conf.layers.samediff.SameDiffLayer;
import org.deeplearning4j.nn.weights.WeightInitUtil;
import org.deeplearning4j.util.CapsuleUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.autodiff.samediff.SDVariable;
import org.nd4j.autodiff.samediff.SameDiff;
import org.nd4j.linalg.api.memory.MemoryWorkspace;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.api.ops.impl.layers.convolution.config.Conv2DConfig;
import org.nd4j.linalg.factory.Nd4j;

@Data
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(buildMethodName = "initBuild", builderMethodName = "innerBuilder")
public class PrimaryCapsules extends SameDiffLayer {

  private static final String WEIGHT_PARAM = "weight";
  private static final String BIAS_PARAM = "bias";
  /**
   * Sets the kernel size of the 2d convolution
   *
   * @param kernelSize
   * @return
   */
  @Builder.Default private int[] kernelSize = new int[] {9, 9};
  /**
   * Sets the stride of the 2d convolution
   *
   * @param stride
   * @return
   */
  @Builder.Default private int[] stride = new int[] {2, 2};
  /**
   * Sets the padding of the 2d convolution
   *
   * @param padding
   * @return
   */
  @Builder.Default private int[] padding = new int[] {0, 0};
  /**
   * Sets the dilation of the 2d convolution
   *
   * @param dilation
   * @return
   */
  @Builder.Default private int[] dilation = new int[] {1, 1};

  private int inputChannels;
  /**
   * Sets the number of channels to use in the 2d convolution.
   *
   * <p>Note that the actual number of channels is channels * capsuleDimensions
   *
   * <p>Does the same thing as nOut()
   *
   * @param channels
   * @return
   */
  @Builder.Default private int channels = 32;

  @Builder.Default private boolean hasBias = true;
  /**
   * Usually inferred automatically.
   *
   * @param capsules
   * @return
   */
  private int capsules;
  /**
   * Sets the number of dimensions to use in the capsules.
   *
   * @param capsuleDimensions
   * @return
   */
  private int capsuleDimensions;
  /**
   * The convolution mode to use in the 2d convolution
   *
   * @param convolutionMode
   * @return
   */
  @Builder.Default private ConvolutionMode convolutionMode = ConvolutionMode.Truncate;
  /**
   * Whether to use a ReLU activation on the 2d convolution
   *
   * @param useRelu
   * @return
   */
  @Builder.Default private boolean useRelU = false;
  /**
   * Use a LeakyReLU activation on the 2d convolution
   *
   * @param leak the alpha value for the LeakyReLU activation.
   * @return
   */
  @Builder.Default private double useLeakyReLU = 0;

  public static PrimaryCapsulesBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static PrimaryCapsulesBuilder<?, ?> builder(
      int capsuleDimensions,
      int channels,
      int[] kernelSize,
      int[] stride,
      int[] padding,
      int[] dilation,
      ConvolutionMode convolutionMode) {
    return innerBuilder()
        .capsuleDimensions(capsuleDimensions)
        .channels(channels)
        .kernelSize(kernelSize)
        .stride(stride)
        .padding(padding)
        .dilation(dilation)
        .convolutionMode(convolutionMode);
  }

  public static PrimaryCapsulesBuilder<?, ?> builder(
      int capsuleDimensions,
      int channels,
      int[] kernelSize,
      int[] stride,
      int[] padding,
      int[] dilation) {
    return innerBuilder()
        .capsuleDimensions(capsuleDimensions)
        .channels(channels)
        .kernelSize(kernelSize)
        .stride(stride)
        .padding(padding)
        .dilation(dilation);
  }

  public static PrimaryCapsulesBuilder<?, ?> builder(
      int capsuleDimensions, int channels, int[] kernelSize, int[] stride, int[] padding) {
    return innerBuilder()
        .capsuleDimensions(capsuleDimensions)
        .channels(channels)
        .kernelSize(kernelSize)
        .stride(stride)
        .padding(padding);
  }

  public static PrimaryCapsulesBuilder<?, ?> builder(
      int capsuleDimensions, int channels, int[] kernelSize, int[] stride) {
    return innerBuilder()
        .capsuleDimensions(capsuleDimensions)
        .channels(channels)
        .kernelSize(kernelSize)
        .stride(stride);
  }

  public static PrimaryCapsulesBuilder<?, ?> builder(
      int capsuleDimensions, int channels, int[] kernelSize) {
    return innerBuilder()
        .capsuleDimensions(capsuleDimensions)
        .channels(channels)
        .kernelSize(kernelSize);
  }

  public static PrimaryCapsulesBuilder<?, ?> builder(int capsuleDimensions, int channels) {
    return innerBuilder().capsuleDimensions(capsuleDimensions).channels(channels);
  }

  @Override
  public SDVariable defineLayer(
      SameDiff SD, SDVariable input, Map<String, SDVariable> paramTable, SDVariable mask) {
    Conv2DConfig conf =
        Conv2DConfig.builder()
            .kH(kernelSize[0])
            .kW(kernelSize[1])
            .sH(stride[0])
            .sW(stride[1])
            .pH(padding[0])
            .pW(padding[1])
            .dH(dilation[0])
            .dW(dilation[1])
            .isSameMode(convolutionMode == ConvolutionMode.Same)
            .build();

    SDVariable conved;

    if (hasBias) {
      conved = SD.cnn.conv2d(input, paramTable.get(WEIGHT_PARAM), paramTable.get(BIAS_PARAM), conf);
    } else {
      conved = SD.cnn.conv2d(input, paramTable.get(WEIGHT_PARAM), conf);
    }

    if (useRelU) {
      if (useLeakyReLU == 0) {
        conved = SD.nn.relu(conved, 0);
      } else {
        conved = SD.nn.leakyRelu(conved, useLeakyReLU);
      }
    }

    SDVariable reshaped = conved.reshape(-1, capsules, capsuleDimensions);
    return CapsuleUtils.squash(SD, reshaped, 2);
  }

  @Override
  public void defineParameters(SDLayerParams params) {
    params.clear();
    params.addWeightParam(
        WEIGHT_PARAM,
        kernelSize[0],
        kernelSize[1],
        inputChannels,
        (long) capsuleDimensions * channels);

    if (hasBias) {
      params.addBiasParam(BIAS_PARAM, (long) capsuleDimensions * channels);
    }
  }

  @Override
  public void initializeParameters(Map<String, INDArray> params) {
    try (MemoryWorkspace ws = Nd4j.getWorkspaceManager().scopeOutOfWorkspaces()) {
      for (Map.Entry<String, INDArray> e : params.entrySet()) {
        if (BIAS_PARAM.equals(e.getKey())) {
          e.getValue().assign(0);
        } else if (WEIGHT_PARAM.equals(e.getKey())) {
          double fanIn = inputChannels * kernelSize[0] * kernelSize[1];
          double fanOut =
              capsuleDimensions
                  * channels
                  * kernelSize[0]
                  * kernelSize[1]
                  / ((double) stride[0] * stride[1]);
          WeightInitUtil.initWeights(
              fanIn, fanOut, e.getValue().shape(), weightInit, null, 'c', e.getValue());
        }
      }
    }
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for Primary Capsules layer (layer name = \""
              + name
              + "\"): expect CNN input.  Got: "
              + inputType);
    }

    if (capsules > 0) {
      return InputType.recurrent(capsules, capsuleDimensions);
    } else {

      InputTypeConvolutional out =
          (InputTypeConvolutional)
              InputTypeUtil.getOutputTypeCnnLayers(
                  inputType,
                  kernelSize,
                  stride,
                  padding,
                  dilation,
                  convolutionMode,
                  (long) capsuleDimensions * channels,
                  -1,
                  getName(),
                  PrimaryCapsules.class);

      return InputType.recurrent(
          (int) (out.getChannels() * out.getHeight() * out.getWidth() / capsuleDimensions),
          capsuleDimensions);
    }
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    if (inputType == null || inputType.getType() != Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for Primary Capsules layer (layer name = \""
              + name
              + "\"): expect CNN input.  Got: "
              + inputType);
    }

    InputTypeConvolutional ci = (InputTypeConvolutional) inputType;

    this.inputChannels = (int) ci.getChannels();

    if (capsules <= 0 || override) {

      InputTypeConvolutional out =
          (InputTypeConvolutional)
              InputTypeUtil.getOutputTypeCnnLayers(
                  inputType,
                  kernelSize,
                  stride,
                  padding,
                  dilation,
                  convolutionMode,
                  (long) capsuleDimensions * channels,
                  -1,
                  getName(),
                  PrimaryCapsules.class);

      this.capsules =
          (int) (out.getChannels() * out.getHeight() * out.getWidth() / capsuleDimensions);
    }
  }

  public static abstract class PrimaryCapsulesBuilder<
          C extends PrimaryCapsules, B extends PrimaryCapsulesBuilder<C, B>>
      extends SameDiffLayerBuilder<C, B> {

    public B kernelSize(int... kernelSize) {
      this.kernelSize$value = ValidationUtils.validate2NonNegative(kernelSize, true, "kernelSize");
      this.kernelSize$set = true;
      return self();
    }

    public B stride(int... stride) {
      this.stride$value = ValidationUtils.validate2NonNegative(stride, true, "stride");
      this.stride$set = true;
      return self();
    }

    public B padding(int... padding) {
      this.padding$value = ValidationUtils.validate2NonNegative(padding, true, "padding");
      this.padding$set = true;
      return self();
    }

    public B dilation(int... dilation) {
      this.dilation$value = ValidationUtils.validate2NonNegative(dilation, true, "dilation");
      this.dilation$set = true;
      return self();
    }
    /**
     * Sets the number of channels to use in the 2d convolution.
     *
     * <p>Note that the actual number of channels is channels * capsuleDimensions
     *
     * <p>Does the same thing as channels()
     *
     * @param nOut
     * @return
     */
    public B nOut(int nOut) {
      return channels(nOut);
    }
    /**
     * Use a ReLU activation on the 2d convolution
     *
     * @return
     */
    public B useReLU() {
      return useRelU(true);
    }

    /**
     * Use a LeakyReLU activation on the 2d convolution. Implies {@link #useReLU()} set true.
     *
     * @param leak the alpha value for the LeakyReLU activation.
     * @return
     */
    public B useLeakyReLU(double leak) {
      this.useRelU(true);
      this.useLeakyReLU$value = leak;
      this.useLeakyReLU$set = true;
      return self();
    }

    public C build() {
      C l = initBuild();
      if (capsuleDimensions <= 0 || channels$value <= 0) {
        throw new IllegalArgumentException(
            "Invalid configuration for Primary Capsules (layer name = \""
                + l.getName()
                + "\"):"
                + " capsuleDimensions and channels must be > 0.  Got: "
                + capsuleDimensions
                + ", "
                + channels$value);
      }

      if (capsules < 0) {
        throw new IllegalArgumentException(
            "Invalid configuration for Capsule ILayer (layer name = \""
                + l.getName()
                + "\"):"
                + " capsules must be >= 0 if set.  Got: "
                + capsules);
      }
      return l;
    }
  }
}
