/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.*;
import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.layers.convolution.SeparableConvolution2DLayer;
import org.deeplearning4j.nn.params.SeparableConvolutionParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

/**
 * SeparableConvolution2D layer nIn in the input layer is the number of channels nOut is the number
 * of filters to be used in the net or in other words the channels The builder specifies the
 * filter/kernel size, the stride and padding The pooling layer takes the kernel size
 */
@Data
@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(buildMethodName = "initBuild", builderMethodName = "innerBuilder")
public class SeparableConvolution2D extends ConvolutionLayer {
  /**
   * Set constraints to be applied to the point-wise convolution weight parameters of this layer.
   * Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
   * regularization, etc). These constraints are applied at each iteration, after the parameters
   * have been updated.
   */
  protected List<LayerConstraint> pointWiseConstraints;
  /**
   * Set channels multiplier of channels-wise step in separable convolution
   *
   * @param depthMultiplier integer value, for each input map we get depthMultipler outputs in
   *     channels-wise step.
   * @return Builder
   */
  @Builder.Default private int depthMultiplier = 1;
  /**
   * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
   * See {@link CNN2DFormat} for more details.<br>
   * Default: NCHW
   *
   * @param format Format for activations (in and out)
   */
  @Builder.Default
  protected CNN2DFormat dataFormat =
          CNN2DFormat.NCHW; // default value for legacy serialization reasons
  public static SeparableConvolution2DBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static SeparableConvolution2DBuilder<?, ?> builder(int... kernelSize) {
    return innerBuilder().kernelSize(kernelSize);
  }

  protected boolean allowCausal() {
    // Causal convolution - allowed for 1D only
    return false;
  }

  @Override
  protected void initializeConstraints() {
    super.initializeConstraints();
    if (pointWiseConstraints != null) {
      if (constraints == null) {
        constraints = new ArrayList<>();
      }
      for (LayerConstraint constraint : pointWiseConstraints) {
        LayerConstraint clonedConstraint = constraint.clone();
        clonedConstraint.setParams(
            Collections.singleton(SeparableConvolutionParamInitializer.POINT_WISE_WEIGHT_KEY));
        constraints.add(clonedConstraint);
      }
    }
  }

  public boolean hasBias() {
    return isHasBias();
  }

  @Override
  public SeparableConvolution2D clone() {
    SeparableConvolution2D clone = (SeparableConvolution2D) super.clone();
    if (clone.getKernelSize() != null) {
      clone.setKernelSize( clone.getKernelSize().clone());
    }
    if (clone.getStride() != null) {
      clone.setStride( clone.getStride().clone());
    }
    if (clone.getPadding() != null) {
      clone.setPadding( clone.getPadding().clone());
    }
    return clone;
  }

  @Override
  public Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerValidation.assertNInNOutSet(
        "SeparableConvolution2D", getName(), layerIndex, getNIn(), getNOut());
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.convolution.SeparableConvolution2DLayer ret =
        new org.deeplearning4j.nn.layers.convolution.SeparableConvolution2DLayer(
            lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);

    return ret;
  }

  @Override
  public ParamInitializer initializer() {
    return SeparableConvolutionParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.CNN) {
      throw new IllegalStateException(
          "Invalid input for Convolution layer (layer name=\""
              + getName()
              + "\"): Expected CNN input, got "
              + inputType);
    }

    CNN2DFormat format = ((InputType.InputTypeConvolutional) inputType).getFormat();

    return InputTypeUtil.getOutputTypeCnnLayers(
        inputType,
        getKernelSize(),
        getStride(),
        getPadding(),
        getDilation(),
        getConvolutionMode(),
        nOut,
        layerIndex,
        getName(),
        format,
        SeparableConvolution2DLayer.class);
  }

  public static abstract class SeparableConvolution2DBuilder<
          C extends SeparableConvolution2D, B extends SeparableConvolution2DBuilder<C, B>>
      extends ConvolutionLayerBuilder<C, B> {
    public C build() {
      C l = this.initBuild();
      if (l.getKernelSize().length != 2) {
        throw new IllegalArgumentException("Kernel size of should be rows x columns (a 2d array)");
      }

      if (l.getStride().length != 2) {
        throw new IllegalArgumentException(
            "Stride should include stride for rows and columns (a 2d array)");
      }

      if (l.getPadding().length != 2) {
        throw new IllegalArgumentException(
            "Padding should include padding for rows and columns (a 2d array)");
      }
      l.initializeConstraints();
      return l;
    }

    /**
     * Set constraints to be applied to the point-wise convolution weight parameters of this layer.
     * Default: no constraints.<br>
     * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
     * regularization, etc). These constraints are applied at each iteration, after the parameters
     * have been updated.
     *
     * @param constraints Constraints to apply to the point-wise convolution parameters of this
     *     layer
     */
    public B constrainPointWise(LayerConstraint... constraints) {
      this.pointWiseConstraints = Arrays.asList(constraints);
      return self();
    }

    @Override
    public B kernelSize(int... kernelSize) {
      super.kernelSize(ValidationUtils.validate2NonNegative(kernelSize, false, "kernelSize"));
      return self();
    }

    @Override
    public B stride(int... stride) {
      super.stride(ValidationUtils.validate2NonNegative(stride, false, "stride"));
      return self();
    }

    @Override
    public B padding(int... padding) {
      super.padding(ValidationUtils.validate2NonNegative(padding, false, "kernelSize"));
      return self();
    }
  }
}
