/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import java.util.Collection;
import java.util.Map;
import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.NoArgsConstructor;
import lombok.ToString;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.RNNFormat;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.Convolution1DUtils;
import org.deeplearning4j.util.ConvolutionUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

/*
 * Currently, we just subclass off the SubsamplingLayer and hard code the "width" dimension to 1.
 * TODO: We will eventually want to NOT subclass off of SubsamplingLayer.
 * This approach treats a multivariate time series with L timesteps and
 * P variables as an L x 1 x P image (L rows high, 1 column wide, P
 * channels deep). The kernel should be H<L pixels high and W=1 pixels
 * wide.
 */

@ToString(callSuper = true)
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(buildMethodName = "initBuild")
public class Subsampling1DLayer extends SubsamplingLayer {

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {
    LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
    runInheritance();
    org.deeplearning4j.nn.layers.convolution.subsampling.Subsampling1DLayer ret =
        new org.deeplearning4j.nn.layers.convolution.subsampling.Subsampling1DLayer(
            lconf, networkDataType);
    ret.addTrainingListeners(trainingListeners);
    ret.setIndex(layerIndex);
    ret.setParamsViewArray(layerParamsView);
    Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
    ret.setParamTable(paramTable);
    ret.setLayerConfiguration(lconf);
    return ret;
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    if (inputType == null || inputType.getType() != InputType.Type.RNN) {
      throw new IllegalStateException(
          "Invalid input for Subsampling1D layer (layer name=\""
              + getName()
              + "\"): Expected RNN input, got "
              + inputType);
    }
    InputType.InputTypeRecurrent r = (InputType.InputTypeRecurrent) inputType;
    long inputTsLength = r.getTimeSeriesLength();
    long outLength;
    if (inputTsLength < 0) {
      // Probably: user did InputType.recurrent(x) without specifying sequence length
      outLength = -1;
    } else {
      outLength =
          Convolution1DUtils.getOutputSize(
              inputTsLength, kernelSize[0], stride[0], padding[0], convolutionMode, getDilation()[0]);
    }
    return InputType.recurrent(r.getSize(), outLength, r.getFormat());
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    // No op: subsampling layer doesn't have nIn value
    if (dataFormat == null || override) {
      if (inputType.getType() == InputType.Type.RNN) {
        InputType.InputTypeRecurrent inputTypeConvolutional =
            (InputType.InputTypeRecurrent) inputType;
        this.dataFormat =
            inputTypeConvolutional.getFormat() == RNNFormat.NCW
                ? CNN2DFormat.NCHW
                : CNN2DFormat.NHWC;

      } else if (inputType.getType() == InputType.Type.CNN) {
        InputType.InputTypeConvolutional inputTypeConvolutional =
            (InputType.InputTypeConvolutional) inputType;
        this.dataFormat = inputTypeConvolutional.getFormat();
      }
    }
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    if (inputType == null) {
      throw new IllegalStateException(
          "Invalid input for Subsampling1D layer (layer name=\""
              + getName()
              + "\"): input is null");
    }

    return InputTypeUtil.getPreprocessorForInputTypeRnnLayers(inputType, RNNFormat.NCW, getName());
  }

  @Override
  public Subsampling1DLayer clone() {
    Subsampling1DLayer clone = (Subsampling1DLayer) super.clone();

    if (clone.kernelSize != null) {
      clone.kernelSize = clone.kernelSize.clone();
    }
    if (clone.stride != null) {
      clone.stride = clone.stride.clone();
    }
    if (clone.padding != null) {
      clone.padding = clone.padding.clone();
    }
    if (clone.getDilation() != null) {
      clone.setDilation( clone.getDilation().clone());
    }
    return clone;
  }

  protected boolean allowCausal() {
    return true;
  }

  public static abstract class Subsampling1DLayerBuilder<C extends Subsampling1DLayer, B extends Subsampling1DLayerBuilder<C, B>> extends
  SubsamplingLayerBuilder<C, B> {

    public C build() {
      C l = this.initBuild();
      if (l.getPoolingType() == org.deeplearning4j.nn.conf.layers.PoolingType.PNORM && l.getPnorm() <= 0) {
        throw new IllegalStateException(
                "Incorrect Subsampling config: p-norm must be set when using PoolingType.PNORM");
      }
      ConvolutionUtils.validateConvolutionModePadding(l.getConvolutionMode(), l.getPadding());
      ConvolutionUtils.validateCnnKernelStridePadding(l.getKernelSize(), l.getStride(), l.getPadding());
      return l;
    }

/**
*
   * @param kernelSize
   * @return
*/
    @Override
    public B kernelSize(int... kernelSize) {
      super.kernelSize( ValidationUtils.validate1NonNegative(kernelSize, "kernelSize")[0]);
      return self();
    }

    @Override
    public B stride(int... stride) {
      super.stride( ValidationUtils.validate1NonNegative(stride, "stride")[0]);
      return self();
    }

    @Override
    public B padding(int... padding) {
      super.padding( ValidationUtils.validate1NonNegative(padding, "padding")[0]);
      return self();
    }
  }
}
