/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.params.EmptyParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;

@Data
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(builderMethodName = "innerBuilder")
public class ZeroPadding3DLayer extends NoParamLayer {
    /**
     * [padLeftD, padRightD, padLeftH, padRightH, padLeftW, padRightW]
     */
    @Builder.Default
    private int[] padding = new int[] {0, 0, 0, 0, 0, 0};// [padLeftD, padRightD, padLeftH, padRightH, padLeftW, padRightW]

    public static ZeroPadding3DLayerBuilder<?,?> builder() {
        return innerBuilder();
    }
    /**
     * @param padding Padding for both the left and right in all three spatial dimensions
     */
    public static ZeroPadding3DLayerBuilder<?,?> builder(int padding) {
        return innerBuilder()
                .padding(padding, padding, padding, padding, padding, padding);
    }
    /**
     * @param padding Padding for both the left and right in all three spatial dimensions
     */
    public static ZeroPadding3DLayerBuilder<?,?> builder(int[] padding) {
        return innerBuilder()
                .padding(padding);
    }
    /**
     * Use same padding for left and right boundaries in depth, height and width.
     *
     * @param padDepth padding used for both depth boundaries
     * @param padHeight padding used for both height boundaries
     * @param padWidth padding used for both width boudaries
     */
    public static ZeroPadding3DLayerBuilder<?,?> builder(int padDepth, int padHeight, int padWidth) {
        return innerBuilder()
                .padding(padDepth, padDepth, padHeight, padHeight, padWidth, padWidth);
    }
    /**
     * Explicit padding of left and right boundaries in depth, height and width dimensions
     *
     * @param padLeftD Depth padding left
     * @param padRightD Depth padding right
     * @param padLeftH Height padding left
     * @param padRightH Height padding right
     * @param padLeftW Width padding left
     * @param padRightW Width padding right
     */
    public static ZeroPadding3DLayerBuilder<?,?> builder(int padLeftD, int padRightD, int padLeftH, int padRightH, int padLeftW, int padRightW) {
        return innerBuilder()
                .padding(padLeftD, padRightD, padLeftH, padRightH, padLeftW, padRightW);
    }
public static abstract class ZeroPadding3DLayerBuilder<C extends ZeroPadding3DLayer, B extends ZeroPadding3DLayerBuilder<C, B>> extends
        NoParamLayerBuilder<C, B> {

    /**
     * [padLeftD, padRightD, padLeftH, padRightH, padLeftW, padRightW]
     */
    public B padding(int... padding) {
        this.padding$value = ValidationUtils.validate6NonNegative(padding, "padding");
        this.padding$set = true;
        return self();
    }
}
    @Override
    public org.deeplearning4j.nn.api.Layer instantiate(NeuralNetConfiguration conf,
                                                       Collection<TrainingListener> iterationListeners, int layerIndex, INDArray layerParamsView,
                                                       boolean initializeParams, DataType networkDataType) {
        LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);

        org.deeplearning4j.nn.layers.convolution.ZeroPadding3DLayer ret =
                        new org.deeplearning4j.nn.layers.convolution.ZeroPadding3DLayer(lconf, networkDataType);
        ret.addTrainingListeners(iterationListeners);
        ret.setIndex(layerIndex);
        Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
        ret.setParamTable(paramTable);
        ret.setLayerConfiguration(lconf);
        return ret;
    }

    @Override
    public ParamInitializer initializer() {
        return EmptyParamInitializer.getInstance();
    }

    @Override
    public InputType getOutputType(int layerIndex, InputType inputType) {
        if (inputType == null || inputType.getType() != InputType.Type.CNN3D) {
            throw new IllegalStateException("Invalid input for 3D CNN layer (layer index = " + layerIndex
                            + ", layer name = \"" + getName() + "\"): expect CNN3D input type with size > 0. Got: "
                            + inputType);
        }
        InputType.InputTypeConvolutional3D c = (InputType.InputTypeConvolutional3D) inputType;
        return InputType.convolutional3D(c.getDepth() + padding[0] + padding[1],
                        c.getHeight() + padding[2] + padding[3], c.getWidth() + padding[4] + padding[5],
                        c.getChannels());
    }

    @Override
    public void setNIn(InputType inputType, boolean override) {
        //No op
    }

    @Override
    public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
        if (inputType == null) {
            throw new IllegalStateException("Invalid input for ZeroPadding3DLayer layer (layer name=\"" + getName()
                            + "\"): input is null");
        }

        return InputTypeUtil.getPreProcessorForInputTypeCnn3DLayers(inputType, getName());
    }

    @Override
    public boolean isPretrainParam(String paramName) {
        throw new UnsupportedOperationException("ZeroPadding3DLayer does not contain parameters");
    }

    @Override
    public LayerMemoryReport getMemoryReport(InputType inputType) {
        InputType outputType = getOutputType(-1, inputType);

        return new LayerMemoryReport.Builder(name, ZeroPadding3DLayer.class, inputType, outputType)
                        .standardMemory(0, 0) //No params
                        .workingMemory(0, 0, MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS)
                        .cacheMemory(MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) //No caching
                        .build();
    }


}
