/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import lombok.*;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.conf.CNN2DFormat;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.deeplearning4j.util.ConvolutionUtils;
import org.deeplearning4j.util.ValidationUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;

@Data
@EqualsAndHashCode(callSuper = true)
@SuperBuilder(builderMethodName = "innerBuilder", buildMethodName = "initBuild")
public class ZeroPaddingLayer extends NoParamLayer {
    /**
     * @param padding Padding value for top, bottom, left, and right. Must be length 4 array
     */
    @Builder.Default
    private int[] padding = new int[] {0, 0, 0, 0}; //Padding: top, bottom, left, right

    public static ZeroPaddingLayerBuilder<?,?> builder() {
        return innerBuilder();
    }

    /**
     * @param padHeight Padding for both the top and bottom
     * @param padWidth Padding for both the left and right
     */
    public static ZeroPaddingLayerBuilder<?,?> builder(int padHeight, int padWidth) {
        return innerBuilder()
                .padding(padHeight, padHeight, padWidth, padWidth);
    }
    /**
     * @param padTop Top padding value
     * @param padBottom Bottom padding value
     * @param padLeft Left padding value
     * @param padRight Right padding value
     */
    public static ZeroPaddingLayerBuilder<?,?> builder(int padTop, int padBottom, int padLeft, int padRight) {
        return innerBuilder()
                .padding(padTop, padBottom, padLeft, padRight);
    }

    public static ZeroPaddingLayerBuilder<?,?> builder(int[] padding) {
        return innerBuilder()
                .padding(padding);
    }

    public static abstract class ZeroPaddingLayerBuilder<C extends ZeroPaddingLayer, B extends ZeroPaddingLayerBuilder<C, B>>
    extends NoParamLayerBuilder<C, B> {
        public C build() {
            if (padding$value == null || padding$value.length != 4) {
                throw new IllegalArgumentException(
                        "Invalid padding values: must have exactly 4 values [top, bottom, left, right]." + " Got: "
                                + (padding$value == null ? null : Arrays.toString(padding$value)));
            }

            C l = initBuild();
            l.initializeConstraints();
            return l;
        }

        public B padding(int ... padding) {
            this.padding$value = ValidationUtils.validate4NonNegative(padding, "padding");
            this.padding$set = true;
            return self();
        }
    }
    /**
     * Set the data format for the CNN activations - NCHW (channels first) or NHWC (channels last).
     * See {@link CNN2DFormat} for more details.<br>
     * Default: NCHW
     * @param format Format for activations (in and out)
     */
    @Builder.Default
    private CNN2DFormat dataFormat = CNN2DFormat.NCHW;



    @Override
    public org.deeplearning4j.nn.api.Layer instantiate(NeuralNetConfiguration conf,
                                                       Collection<TrainingListener> trainingListeners, int layerIndex, INDArray layerParamsView,
                                                       boolean initializeParams, DataType networkDataType) {
        LayerConfiguration lconf = conf.getFlattenedLayerConfigurations().get(layerIndex);
runInheritance();
        org.deeplearning4j.nn.layers.convolution.ZeroPaddingLayer ret =
                        new org.deeplearning4j.nn.layers.convolution.ZeroPaddingLayer(lconf, networkDataType);
        ret.addTrainingListeners(trainingListeners);
        ret.setIndex(layerIndex);
        Map<String, INDArray> paramTable = initializer().init(this, layerParamsView, initializeParams);
        ret.setParamTable(paramTable);
        ret.setLayerConfiguration(lconf);
        return ret;
    }

    @Override
    public InputType getOutputType(int layerIndex, InputType inputType) {
        int[] hwd = ConvolutionUtils.getHWDFromInputType(inputType);
        int outH = hwd[0] + padding[0] + padding[1];
        int outW = hwd[1] + padding[2] + padding[3];

        InputType.InputTypeConvolutional c = (InputType.InputTypeConvolutional)inputType;

        return InputType.convolutional(outH, outW, hwd[2], c.getFormat());
    }

    @Override
    public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
        Preconditions.checkArgument(inputType != null, "Invalid input for ZeroPaddingLayer layer (layer name=\""
                        + getName() + "\"): InputType is null");
        return InputTypeUtil.getPreProcessorForInputTypeCnnLayers(inputType, getName());
    }

    @Override
    public LayerMemoryReport getMemoryReport(InputType inputType) {
        InputType outputType = getOutputType(-1, inputType);

        return new LayerMemoryReport.Builder(name, ZeroPaddingLayer.class, inputType, outputType)
                        .standardMemory(0, 0) //No params
                        //Inference and training is same - just output activations, no working memory in addition to that
                        .workingMemory(0, 0, MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS)
                        .cacheMemory(MemoryReport.CACHE_MODE_ALL_ZEROS, MemoryReport.CACHE_MODE_ALL_ZEROS) //No caching
                        .build();
    }

    @Override
    public void setNIn(InputType inputType, boolean override) {
        InputType.InputTypeConvolutional c = (InputType.InputTypeConvolutional)inputType;
        this.dataFormat = c.getFormat();
    }


}
