/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers.misc;

import java.util.Collection;
import java.util.List;
import java.util.Set;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.Setter;
import lombok.experimental.SuperBuilder;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.params.FrozenLayerParamInitializer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.regularization.Regularization;

@EqualsAndHashCode(callSuper = false)
@SuperBuilder(builderMethodName = "innerBuilder")
public class FrozenLayer extends LayerConfiguration {

  /** A layer configuration, only if this layer config has been created from another one */
  @Getter @Setter private LayerConfiguration innerConfiguration;

  public static FrozenLayerBuilder<?, ?> builder() {
    return innerBuilder();
  }

  public static FrozenLayerBuilder<?, ?> builder(LayerConfiguration innerConfiguration) {
    return innerBuilder().innerConfiguration(innerConfiguration);
  }

  @Override
  public LayerConfiguration clone() {
    FrozenLayer l = (FrozenLayer) super.clone();
    l.innerConfiguration = innerConfiguration.clone();
    return l;
  }

  @Override
  public org.deeplearning4j.nn.api.Layer instantiate(
      NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType) {

    // Need to be able to instantiate a layer, from a config - for JSON -> net type situations
    org.deeplearning4j.nn.api.Layer underlying =
        innerConfiguration.instantiate(
            getNetConfiguration(),
            trainingListeners,
            layerIndex,
            layerParamsView,
            initializeParams,
            networkDataType);

    NeuralNetConfiguration nncUnderlying = underlying.getNetConfiguration();
    if (nncUnderlying.getNetWideVariables() != null) {
      Set<String> vars = nncUnderlying.getNetWideVariables(true);
      nncUnderlying.clearNetWideVariable();
      conf.clearNetWideVariable();
      for (String s : vars) {
        conf.getNetWideVariables(false).add(s);
        nncUnderlying.getNetWideVariables(false).add(s);
      }
    }

    return new org.deeplearning4j.nn.layers.FrozenLayer(underlying);
  }

  @Override
  public ParamInitializer initializer() {
    return FrozenLayerParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    return innerConfiguration.getOutputType(layerIndex, inputType);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    innerConfiguration.setNIn(inputType, override);
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    return innerConfiguration.getPreProcessorForInputType(inputType);
  }

  @Override
  public List<Regularization> getRegularizationByParam(String param) {
    return null;
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    return false;
  }

  @Override
  public IUpdater getUpdaterByParam(String paramName) {
    return null;
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    return innerConfiguration.getMemoryReport(inputType);
  }

  @Override
  public void setName(String layerName) {
    super.setName(layerName);
    innerConfiguration.setName(layerName);
  }

  @Override
  public void setConstraints(List<LayerConstraint> constraints) {
    this.constraints = constraints;
    this.innerConfiguration.setConstraints(constraints);
  }
}
